@@ -1,5 +1,5 @@
 import hmac as hmaca
-from base64 import b64encode, b64decode
+from base64 import b64decode, b64encode
 
 
 def hmac(hf, key, msg):
